# Description: A set of fetch methods to get macroeconomic data

# Dependencies
from fredapi import Fred
import pandas as pd
from datetime import datetime, timedelta
import json
import pickle
# Used for downloading data from the internet
import requests
import zipfile
import tempfile
import shutil

# Get current directory
import sys, os
CWD = os.path.dirname(os.path.dirname(__file__))
sys.path.append(CWD)

# Package imports
import timeseries as mt

# Directory for data cache
CACHE_DIR = os.path.join(os.path.expanduser('~'), '.macro_cache')

def parse_year(year: str) -> datetime:
    '''
    Get year, month, day from %Y.%f
    '''
    year_int = int(year)
    month = int((year % 1) * 12) + 1
    day = int(((year % 1) * 12) % 1 * 30) + 1 
    return datetime(year_int, month, day)

def get_fred(data_sources: list[str], data_names: list[str], start_date: str = "", end_date: str = "", api_key=None, force_fetch=False):
    '''
    Get data from FRED.

    Parameters
    ----------
    data_sources : list[str]
        List of FRED series IDs.
    data_names : list[str]
        List of names to assign to the data.
    date_one : str, optional
        Start date for data. Default is "".
    date_two : str, optional
        End date for data. Default is "".
    api_key : str, optional
        API key for FRED. Default is None (not needed if using cached data).

    Returns
    -------
    mt.TimeseriesTable
        Table of FRED data.
    
    Notes
    -----
    - Data is automatically set to year/quarter/month-end frequency and reindexed daily.
    - Caching is used to store data for one week.
    '''
    # Look for api key 
    if api_key is None: # Load API key
        try:
            with open(f'{os.path.dirname(__file__)}/fred_api_key.txt') as f:
                api_key = f.read().strip()
        except FileNotFoundError as e:
            try: 
                api_key = os.environ['FRED_API_KEY']
            except KeyError as e:
                pass # api_key is None
    
    # Create cache directory
    cache_dir = os.path.join(CACHE_DIR, 'fred_data')
    if not os.path.exists(cache_dir):
        print(f"Cache directory does not exist. Creating cache directory at {cache_dir}")
        os.makedirs(cache_dir)

    # Create connection to fred 
    if api_key:
        fred_connection = Fred(api_key=api_key)
    else:
        fred_connection = None
            
    # Get data from FRED
    data = mt.TimeseriesTable()
    for idx, series_id in enumerate(data_sources):
        # map series_id to data_name
        data_name = data_names[idx]

        # Check if data is in cache
        cache_file = os.path.join(cache_dir, f"{series_id}.pkl")
        metadata_file = os.path.join(cache_dir, f"{series_id}_metadata.json")
        if os.path.exists(cache_file) and os.path.exists(metadata_file):
            with open(metadata_file, 'r') as f:
                metadata = json.load(f)
                cache_expiry = datetime.strptime(metadata['expiry_date'], '%Y-%m-%d')
                cache_start_date = metadata['start_date']
                cache_end_date = metadata['end_date']
                # If end_date is not provided then query the entire series if it's been more than a day
                if force_fetch or (end_date == "" and datetime.now() - cache_expiry > timedelta(days=1)):
                    data_is_fresh = False
                else:
                    data_is_fresh = True
                # Make sure the cache is still valid 
                if datetime.now() < cache_expiry and start_date >= cache_start_date and end_date <= cache_end_date and data_is_fresh:
                    with open(cache_file, 'rb') as f:
                        cached_data = pickle.load(f)
                    data[data_name] = cached_data
                    print(f"Data for {data_name} loaded from cache.")
                    continue
        
        # Get data from FRED (cache is expired or invalid)
        if fred_connection:
            raw_data = fred_connection.get_series(series_id, observation_start=start_date, observation_end=end_date)
            data_freq = fred_connection.get_series_info(series_id)['frequency']
        # Get data from FRED web page
        else: 
            # Add deprecation warning
            print("DeprecationWarning: Fetching data from FRED without an API key is deprecated. Please provide an API key to avoid this warning.", DeprecationWarning)
            page_url = f'https://fred.stlouisfed.org/series/{series_id}/downloaddata/{series_id}.csv'
            try:
                raw_data = pd.read_csv(page_url, index_col='DATE', parse_dates=True)
                # case to pd.Series
                raw_data = pd.Series(raw_data.squeeze())
            except Exception as e:
                raise ValueError(f"Error fetching data from FRED: {e}")
            data_freq = 'unknown'
        new_data = mt.Timeseries(raw_data, name=data_name, source_freq=data_freq, data_source="FRED")
        new_data.transformations.append('reindex_daily_end')
        print(f"Data for {data_name} fetched from FRED.")
        data[data_name] = new_data

        # Save data to cache
        with open(cache_file, 'wb') as f:
            pickle.dump(new_data, f)
        metadata = {
            'created_date': datetime.now().strftime('%Y-%m-%d'),
            'expiry_date': (datetime.now() + timedelta(weeks=1)).strftime('%Y-%m-%d'),
            'start_date': start_date,
            'end_date': end_date
        }
        with open(metadata_file, 'w') as f:
            json.dump(metadata, f)

    print('NOTE: FRED data is automatically set to year/quarter/month-end frequency and reindexed daily.\n')
    return data

def get_barnichon(filepath: str, table: mt.TimeseriesTable, input_var: str, output_name: str):
    '''
    Get vacancy rate data from Barnichon.

    Parameters
    ----------
    filepath : str
        Path to Barnichon file.
    table : mt.TimeseriesTable
        Table to store data.
    input_var : str
        Input variable to use.
    output_name : str
        Name to assign to the output variable.

    Returns
    -------
    mt.TimeseriesTable
        Table with vacancy rate data.

    Notes
    -----
    - Data is automatically set to quarterly frequency.
    - Necessary data is automatically fetched from FRED if needed.
    '''
    # read data from barnichon file
    data = pd.read_csv(filepath)
    data['year'] = data['year'].apply(parse_year)
    data.set_index('year', inplace=True)
    data = mt.TimeseriesTable(data)
    # data = data._update(data.ffill())

    match input_var: # construct data
        case 'V_LF':
            if 'JTSJOL' not in table:
                table['V'] = get_fred(['JTSJOL'], ['V'])['V']
            if 'CLF16OV' not in table:
                table['L'] = get_fred(['CLF16OV'], ['L'])['L']
            # Create vacancy rate with JOLTS data
            table[output_name] = mt.Timeseries(100 * table['V'] / table['L'], name=output_name, source_freq='Q', data_source='Barnichon')
            # Replace pre-2001 with Barnichon's vacancy rate
            barnichon_data = data.df[:'2000-12-31'][input_var]
            for idx in barnichon_data.index:
                if idx < datetime(2001, 1, 1):
                    table[output_name][idx] = barnichon_data[idx]
        case 'V_hwi':
            tr_all = (data.index[0], data.index[-1])
            table[output_name] = mt.Timeseries(data[input_var], name=output_name, source_freq='Q', data_source='Barnichon')
        case _:
            raise ValueError(f'Unknown input_var: {input_var}')
        
    return table

def get_ludvigson():
    '''
    Get macroeconomics uncertaintity data from Ludvigson.

    Returns
    -------
    dict
        Dictionary of TimeseriesTables for each data source (FinancialUncertainty, MacroUncertainty, RealUncertainty).

    Notes
    -----
    - Data is automatically set to monthly frequency.
    - Caching is used to store data for one week.
    '''
    # Define data source
    DATA_SOURCE = 'https://www.sydneyludvigson.com/s/MacroFinanceUncertainty_202408Update.zip'
    
    # Find the cache 
    cache_dir = os.path.join(CACHE_DIR, 'ludvigson_data')
    if not os.path.exists(cache_dir):
        print(f"Cache directory does not exist. Creating cache directory at {cache_dir}")
        os.makedirs(cache_dir)

    # Check if data is in cache
    data_files = ['FinancialUncertainty', 'MacroUncertainty', 'RealUncertainty']
    data = {}
    for data_file in data_files:
        cache_file = os.path.join(cache_dir, f"{data_file}.pkl")
        metadata_file = os.path.join(cache_dir, f"{data_file}_metadata.json")
        if os.path.exists(cache_file) and os.path.exists(metadata_file):
            with open(metadata_file, 'r') as f:
                metadata = json.load(f)
                cache_expiry = datetime.strptime(metadata['expiry_date'], '%Y-%m-%d')
                # Make sure the cache is still valid
                if datetime.now() < cache_expiry:
                    with open(cache_file, 'rb') as f:
                        cached_data = pickle.load(f)
                        data[data_file] = cached_data
                    print(f"Data for {data_file} loaded from cache.")
                    continue
    
    # Check if all data is in cache
    if len(data) == len(data_files):
        return data
    
    # Clear old files from the cache
    shutil.rmtree(cache_dir)
    os.makedirs(cache_dir)
    # Download data from Ludvigson
    response = requests.get(DATA_SOURCE)
    zip_path = os.path.join(tempfile.gettempdir(), 'temp.zip')
    # Save the zip file to a temporary file
    with open(zip_path, 'wb') as f:
        f.write(response.content)
    # Extract the contents of the zip file
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(cache_dir)

    # Parse the contents of the zip file
    file_names = os.listdir(cache_dir)
    for source_file in file_names:
        # Get the data
        data_name = [elem for elem in data_files if elem in source_file][0]
        data_path = os.path.join(cache_dir, source_file)
        src_data = pd.read_excel(data_path)
        # Construct TimeseriesTable
        table = mt.TimeseriesTable()
        # Set index
        src_data.set_index('Date', inplace=True)
        # Create timeseries
        for col in src_data.columns:
            table[col] = mt.Timeseries(src_data[col], name=col, source_freq='monthly', data_source='Ludvigson')
        data[data_name] = table
        # Save data to cache
        cache_file = os.path.join(cache_dir, f"{data_name}.pkl")
        with open(cache_file, 'wb') as f:
            pickle.dump(table, f)
        metadata = {
            'created_date': datetime.now().strftime('%Y-%m-%d'),
            'expiry_date': (datetime.now() + timedelta(weeks=1)).strftime('%Y-%m-%d')
        }
        metadata_file = os.path.join(cache_dir, f"{data_name}_metadata.json")
        with open(metadata_file, 'w') as f:
            json.dump(metadata, f)
        print(f"Data for {data_name} fetched from Ludvigson.")
    
    return data



import unittest
import pandas as pd
import numpy as np
from datetime import datetime
from src.macroecon_tools.timeseries import Timeseries

# FILE: Python/macroecon-tools/src/macroecon_tools/test_timeseries.py


class TestTimeseries(unittest.TestCase):

    def setUp(self):
        # Create sample data for Timeseries
        self.sample_data = pd.Series([i for i in range(24)], index=pd.date_range("2020-01-01", periods=24, freq="ME"))
        self.ts = Timeseries(self.sample_data, name='ts', source_freq='D')

    def test_initialization(self):
        self.assertIsInstance(self.ts, Timeseries)
        self.assertEqual(self.ts.name, 'ts')
        self.assertEqual(self.ts.source_freq, 'D')
        self.assertEqual(len(self.ts), 100)

    def test_logdiff(self):
        transformed_ts = self.ts.logdiff(1)
        self.assertIsInstance(transformed_ts, Timeseries)
        self.assertIn('logdiff_1_D', transformed_ts.transformations)

    def test_diff(self):
        transformed_ts = self.ts.diff(1)
        self.assertIsInstance(transformed_ts, Timeseries)
        self.assertIn('diff_1', transformed_ts.transformations)

    def test_log(self):
        transformed_ts = self.ts.log()
        self.assertIsInstance(transformed_ts, Timeseries)
        self.assertIn('log', transformed_ts.transformations)

    def test_log100(self):
        transformed_ts = self.ts.log100()
        self.assertIsInstance(transformed_ts, Timeseries)
        self.assertIn('log100', transformed_ts.transformations)

    def test_agg(self):
        aggregated_ts = self.ts.agg('monthly', 'mean')
        self.assertIsInstance(aggregated_ts, Timeseries)
        self.assertIn('agg_monthly_mean', aggregated_ts.transformations)

    def test_trunc(self):
        truncated_ts = self.ts.truncate('2020-01-10', '2020-01-20')
        self.assertIsInstance(truncated_ts, Timeseries)
        self.assertEqual(len(truncated_ts), 11)
        self.assertEqual(truncated_ts.index[0], pd.Timestamp('2020-01-10'))
        self.assertEqual(truncated_ts.index[-1], pd.Timestamp('2020-01-20'))

    def test_dropna(self):
        self.ts.iloc[0] = np.nan
        dropped_ts = self.ts.dropna()
        self.assertIsInstance(dropped_ts, Timeseries)
        self.assertEqual(len(dropped_ts), 99)

    def test_linear_filter(self):
        filtered_ts = self.ts.linear_filter('01-Jan-2020', '10-Jan-2020')
        self.assertIsInstance(filtered_ts, Timeseries)
        self.assertIn('linear_filter_2020-01-01_2020-01-10', filtered_ts.transformations)

    def test_hamilton_filter(self):
        filtered_ts = self.ts.hamilton_filter('01-Jan-2020', '10-Jan-2020')
        self.assertIsInstance(filtered_ts, Timeseries)
        self.assertIn('hamilton_filter_2020-01-01_2020-01-10_12_24', filtered_ts.transformations)

    def test_save_load(self):
        file_path = 'test_timeseries.pkl'
        self.ts.save(file_path)
        loaded_ts = Timeseries([]).load(file_path)
        self.assertEqual(len(loaded_ts), 100)
        self.assertEqual(loaded_ts.name, 'ts')
        self.assertEqual(loaded_ts.source_freq, 'D')

if __name__ == '__main__':
    unittest.main()


# Description: a visulization module for timeseries data

# Import visualization libraries
import matplotlib.pyplot as plt

# Add path to current directory
import sys, os
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

# Import timeseries datastructures
from timeseries import Timeseries, TimeseriesTable

# Function to plot multiple timeseries in one graph 
def vis_multi_series(data: TimeseriesTable, save_path: str, variables: dict[str, str] = {}, start_date: str = "", end_date: str = "", is_percent: bool = False):
    """
    Creates a plot of multiple timeseries data in one graph with the data and variables provided.

    Parameters
    ----------
    data : TimeseriesTable
        The timeseries data to be plotted.
    save_path : str
        The path to save the plot, also indicates the type of file to save as (supports pdf, png, eps, etc.)
    variables : dict[str, str], optional
        A dictionary containing the variable name and the title of the plot. 
        Default: {} (indicates all variables in the TimeseriesTable will be plotted).
    start_date : str, optional
        The start date of the plot.
        Default: "" (indicates the start date of the data).
    end_date : str, optional
        The end date of the plot. 
        Default: "" (indicates the end date of the data).
    is_percent : bool, optional
        Indicates if the y-axis is in percentage format.
        Default: False (indicates the y-axis is not in percentage format).

    Notes
    -----
    The plot will be saved to the path specified in the save_path parameter.
    """
    # Check for variables
    if len(variables) == 0:
        for var in data:
            variables[var] = var
    # Map variables to list
    variables = [(var, variables[var]) for var in variables]
    # Create a figure and axis
    fig, axs = plt.subplots(len(variables), figsize=(6.5, 4))
    # Plot variables
    for idx, ax in enumerate(axs.flat):
        # plot data
        ax.plot(data.df[variables[idx][0]][start_date:end_date])
        # set title
        ax.set_title(variables[idx][1])
        # set y-axis format
        if is_percent:
            ax.yaxis.set_major_formatter('{x:.0f}%')
        else:
            ax.yaxis.set_major_formatter('{x:.0f}')
        # format graph
        ax.grid()
        ax.autoscale(tight=True)
        ax.label_outer()

    # Use tight layout
    plt.tight_layout()
    # Add 5% padding to each y-axis
    for idx, ax in enumerate(axs.flat):
        ax.margins(y=0.05)

    # Save plot
    plt.savefig(save_path)
    # Close plot
    plt.close()

# Function to plot two variables in one graph
def vis_two_vars(data: TimeseriesTable, x_var: str, y_var: str, save_path: str, title: str = "", start_date: str = "", end_date: str = "", x_is_percent: bool = False, y_is_percent: bool = False):
    """
    Plot two variables in one graph with the data and variables provided.

    Parameters
    ----------
    data : TimeseriesTable
        The timeseries data to be plotted.
    x_var : str
        The variable to be plotted on the x-axis.
    y_var : str
        The variable to be plotted on the y-axis.
    save_path : str
        The path to save the plot, also indicates the type of file to save as (supports pdf, png, eps, etc.)
    title : str, optional
        The title of the plot.
        Default: "" (indicates {x_var} vs {y_var}).
    start_date : str, optional
        The start date of the plot.
        Default: "" (indicates the start date of the data).
    end_date : str, optional
        The end date of the plot. 
        Default: "" (indicates the end date of the data).
    x_is_percent : bool, optional
        Indicates if the x-axis is in percentage format.
        Default: False (indicates the x-axis is not in percentage format).
    y_is_percent : bool, optional
        Indicates if the y-axis is in percentage format.
        Default: False (indicates the y-axis is not in percentage format).
    """
    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(6.5, 4))
    # get start and end date
    if start_date == "":
        # get the start date of the two variables
        start_date = max(data.df[x_var].index[0], data.df[y_var].index[0])
    if end_date == "":
        # get the end date of the two variables
        end_date = min(data.df[x_var].index[-1], data.df[y_var].index[-1])
    # plot data
    ax.scatter(data.df[x_var][start_date:end_date], data.df[y_var][start_date:end_date])
    # set title
    if title == "":
        title = f"{x_var} vs {y_var}"
    ax.set_title(title if title != "" else f"{x_var} vs {y_var}")
    # set axis labels
    ax.set_xlabel(x_var)
    ax.set_ylabel(y_var)
    # set x-axis format
    if x_is_percent:
        ax.xaxis.set_major_formatter('{x:.0f}%')
    else:
        ax.xaxis.set_major_formatter('{x:.0f}')
    # set y-axis format
    if y_is_percent:
        ax.yaxis.set_major_formatter('{x:.0f}%')
    else:
        ax.yaxis.set_major_formatter('{x:.0f}')
    # format graph
    ax.grid()
    ax.autoscale(tight=True)
    ax.label_outer()

    # Save plot
    plt.savefig(save_path)
    # Close plot
    plt.close()

# Visualize multiples variables in one graph
def vis_multi_lines(data: TimeseriesTable, vars: list[str], save_path: str, title: str, start_date: str = "", end_date: str = "", is_percent: bool = False):
    """
    Visualizes multiple variables in one graph with the data and variables provided.

    Parameters
    ----------
    data : TimeseriesTable
        The timeseries data to be plotted.
    vars : list[str]
        The list of variables to be plotted.
    save_path : str
        The path to save the plot, also indicates the type of file to save as (supports pdf, png, eps, etc.)
    title : str
        The title of the plot.
    start_date : str, optional
        The start date of the plot.
        Default: "" (indicates the start date of the data).
    end_date : str, optional
        The end date of the plot. 
        Default: "" (indicates the end date of the data).
    is_percent : bool, optional
        Indicates if the y-axis is in percentage format.
        Default: False (indicates the y-axis is not in percentage format).
    """
    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(6.5, 4))

    # get start and end date
    if start_date == "":
        # get the start date of the two variables
        start_date = min([data.df[var].index[0] for var in vars])
    if end_date == "":
        # get the end date of the two variables
        end_date = max([data.df[var].index[-1] for var in vars])

    # define a list of color and line types to use
    line_types = [('red', '-'), ('blue', '--'), ('darkgreen', '-.'), ('black', ':')]

    # plot data, use different colors and line types if less than 4 variables
    if len(vars) <= len(line_types):
        for idx, var in enumerate(vars):
            ax.plot(data.df[var][start_date:end_date], color=line_types[idx][0], linestyle=line_types[idx][1], label=var)
    else:
        for idx, var in enumerate(vars):
            ax.plot(data.df[var][start_date:end_date], label=var)

    # set title
    ax.set_title(title)
    # set axis labels
    ax.set_xlabel("Date")
    ax.set_ylabel("Value")
    # set y-axis format
    if is_percent:
        ax.yaxis.set_major_formatter('{x:.0f}%')
    else:
        ax.yaxis.set_major_formatter('{x:.0f}')
    # format graph
    ax.grid()
    ax.autoscale(tight=True)
    ax.label_outer()
    # add legend
    ax.legend()

    # Save plot
    plt.savefig(save_path)
    # Close plot
    plt.close()



# Get current directory
# import os, sys
# dir_path = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(dir_path)

# import submodules
from .timeseries import TimeseriesTable, Timeseries
from .fetch_data import get_fred, get_barnichon, get_ludvigson
from .visualizer import vis_multi_series


class Constants:
    """
    A class to hold constant mappings and scales used throughout the application.

    Attributes
    ----------
    freq_map : dict
        A dictionary mapping time frequencies to their corresponding pandas resample codes.
        Keys are:
            'quarterly' -> 'Q'
            'monthly' -> 'M'
            'yearly' -> 'Y'
    agg_map : dict
        A dictionary mapping aggregation methods to their corresponding pandas aggregation functions.
        Keys are:
            'lastvalue' -> 'last'
            'mean' -> 'mean'
            'sum' -> 'sum'
            'min' -> 'min'
            'max' -> 'max'
    ANNSCALE_MAP : dict
        A dictionary mapping time frequencies to their corresponding annualization scales.
        Keys are:
            'daily' -> 36500
            'weekly' -> 5200
            'monthly' -> 1200
            'quarterly' -> 400
            'yearly' -> 100
            'annual' -> 100
    """
    # MATLAB retime map to pandas resample
    freq_map = {
        'hourly': 'H',
        'daily': 'D',
        'weekly': 'W',
        'monthly': 'ME',
        'quarterly': 'QE-DEC',
        'yearly': 'YE',
        'H' : 'H',
        'D' : 'D',
        'W' : 'W-SUN',
        'W-SUN': 'W-SUN',
        'M' : 'ME',
        'ME': 'ME',
        'Q' : 'QE-DEC',
        'QE': 'QE-DEC',
        'QE-DEC': 'QE-DEC',
        'Y' : 'YE-DEC',
        'YE': 'YE-DEC',
        'YE-DEC': 'YE-DEC',
    }
    freq_to_n = ['H', 'D', 'M', 'QE-DEC', 'Y']
    agg_map = {
        'lastvalue': 'last',
        'mean': 'mean',
        'sum': 'sum',
        'min': 'min',
        'max': 'max'
    }

    # Annualization scales
    ANNSCALE_MAP = {
        'daily': 36500,
        'weekly' : 5200,
        'monthly' : 1200,
        'quarterly' : 400,
        'yearly' : 100,
        "annual": 100,
        'D': 36500,
        'W': 5200,
        'M': 1200,
        'Q': 400,
        'QE': 400,
        'QE-DEC': 400,
        'Y': 100
    }

    # group similar date formats
    year_like = ['Y', 'YE', 'yearly', 'annual']
    quarter_like = ['Q', 'QE', 'QE-DEC', 'quarterly']
    month_like = ['M', 'ME', 'monthly']